# Copyright 2015 0xc0170
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import subprocess
import shutil
import logging
import xmltodict
import copy
import re
from codecs import open

from os import getcwd
from os.path import basename, join, normpath
from collections import OrderedDict
from project_generator_definitions.definitions import ProGenDef

from .tool import Tool, Builder, Exporter
from ..util import SOURCE_KEYS

logger = logging.getLogger('progen.tools.uvision')

class uVisionDefinitions():

    debuggers = {
        'ulink2-me': {
            'uvproj': {
                'TargetDlls': {
                    'Driver': 'BIN\\UL2CM3.dll',
                },
                'Utilities': {
                    'Flash2': 'BIN\\UL2CM3.DLL',
                },
            },
            'uvoptx' : {
                'DebugOpt' : {
                    'nTsel' : '1',
                    'pMon': 'BIN\\UL2CM3.DLL',
                },
                'SetRegEntry' : {
                    'Key' : 'UL2CM3',
                },
            },
        },
        'cmsis-dap': {
            'uvproj': {
                'TargetDlls': {
                    'Driver': 'BIN\\CMSIS_AGDI.dll',
                },
                'Utilities': {
                    'Flash2': 'BIN\\CMSIS_AGDI.dll',
                },
            },
            'uvoptx' : {
                'DebugOpt' : {
                    'nTsel' : '12',
                    'pMon': 'BIN\\CMSIS_AGDI.dll',
                },
                'SetRegEntry' : {
                    'Key' : 'CMSIS_AGDI',
                },
            },
        },
        'j-link': {
            'uvproj': {
                'TargetDlls': {
                    'Driver': 'Segger\\JL2CM3.dll',
                },
                'Utilities': {
                    'Flash2': 'Segger\\JL2CM3.dll',
                },
            },
            'uvoptx' : {
                'DebugOpt' : {
                    'nTsel' : '6',
                    'pMon': 'Segger\\JL2CM3.dll',
                },
                'SetRegEntry' : {
                    'Key' : 'JL2CM3',
                },
            },
        },
        'ulink-pro': {
            'uvproj': {
                'TargetDlls': {
                    'Driver': 'BIN\\ULP2CM3.dll',
                },
                'Utilities': {
                    'Flash2': 'BIN\\ULP2CM3.dll',
                },
            },
            'uvoptx' : {
                'DebugOpt' : {
                    'nTsel' : '7',
                    'pMon': 'BIN\\ULP2CM3.DLL',
                },
                'SetRegEntry' : {
                    'Key' : 'ULP2CM3',
                },
            },
        },
        'st-link': {
            'uvproj': {
                'TargetDlls': {
                    'Driver': 'STLink\\ST-LINKIII-KEIL_SWO.dll',
                },
                'Utilities': {
                    'Flash2': 'STLink\\ST-LINKIII-KEIL_SWO.dll',
                },
            },
            'uvoptx' : {
                'DebugOpt' : {
                    'nTsel' : '11',
                    'pMon': 'STLink\\ST-LINKIII-KEIL_SWO.dll',
                },
                'SetRegEntry' : {
                    'Key' : 'ST-LINKIII-KEIL_SWO',
                },
            },
        },
        'nu-link': {
            'uvproj': {
                'TargetDlls': {
                    'Driver': 'BIN\\Nu_Link.dll',
                },
                'Utilities': {
                    'Flash2': 'BIN\\Nu_Link.dll',
                },
            },
            'uvoptx' : {
                'DebugOpt' : {
                    'nTsel' : '9',
                    'pMon': 'NULink\\Nu_Link.dll',
                },
                'SetRegEntry' : {
                    'Key' : 'Nu_Link',
                },
            },
        },
    }

    # use cmsis-dap debugger as default
    debuggers_default = 'cmsis-dap'


class Uvision(Tool, Builder, Exporter):

    optimization_options = ['O0', 'O1', 'O2', 'O3']
    file_types = {'cpp': 8, 'c': 1, 's': 2, 'obj': 3,'o':3, 'lib': 4, 'ar': 4, 'h': 5}

    # flags mapping to uvision uvproj dics
    # for available flags, check armcc/armasm/armlink command line guide
    # this does not provide all options within a project, most usable options are
    # exposed via command line, the rest is covered via template project files
    FLAGS_TO_UVISION = {
        'asm_flags': 'Aads',
        'c_flags': 'Cads',
        'cxx_flags': 'Cads',
        'ld_flags':  'LDads',
    }

    ERRORLEVEL = {
        0: 'success (0 warnings, 0 errors)',
        1: 'warnings',
        2: 'errors',
        3: 'fatal errors',
        11: 'cant write to project file',
        12: 'device error',
        13: 'error writing',
        15: 'error reading xml file',
    }

    SUCCESSVALUE = 0
    WARNVALUE = 1

    generated_project = {
        'path': '',
        'files': {
            'uvproj': '',
        }
    }

    def __init__(self, workspace, env_settings):
        self.definitions = uVisionDefinitions()
        # workspace or project
        self.workspace = workspace
        self.env_settings = env_settings
        self.uvproj_file = join(self.TEMPLATE_DIR, "uvision.uvproj")
        self.uvmpw_file = join(self.TEMPLATE_DIR, "uvision.uvmpw")
        self.uvoptx_file = join(self.TEMPLATE_DIR, "uvision.uvoptx")

    @staticmethod
    def get_toolnames():
        return ['uvision']

    @staticmethod
    def get_toolchain():
        return 'uvision'

    def _expand_one_file(self, source, new_data, extension):
        ordered = OrderedDict()
        ordered["FileType"] = self.file_types[extension]
        ordered["FileName"] = basename(source)
        ordered["FilePath"] = source
        return ordered

    def _normalize_mcu_def(self, mcu_def):
        for k, v in mcu_def['TargetOption'].items():
            mcu_def['TargetOption'][k] = v[0]

    def _uvproj_clean_xmldict(self, uvproj_dic):
        for k, v in uvproj_dic.items():
            if v is None:
                uvproj_dic[k] = ''

    def _uvproj_set_CommonProperty(self, uvproj_dic, project_dic):
        self._uvproj_clean_xmldict(uvproj_dic)

    def _uvproj_set_DebugOption(self, uvproj_dic, project_dic):
        self._uvproj_clean_xmldict(uvproj_dic)
        self._uvproj_clean_xmldict(uvproj_dic['SimDlls'])
        self._uvproj_clean_xmldict(uvproj_dic['Simulator'])
        self._uvproj_clean_xmldict(uvproj_dic['Target'])
        self._uvproj_clean_xmldict(uvproj_dic['TargetDlls'])

    def _uvproj_set_DllOption(self, uvproj_dic, project_dic):
        self._uvproj_clean_xmldict(uvproj_dic)

    def _uvproj_set_TargetArmAds(self, uvproj_dic, project_dic):
        self._uvproj_clean_xmldict(uvproj_dic['Aads'])
        self._uvproj_clean_xmldict(uvproj_dic['Aads']['VariousControls'])
        self._uvproj_clean_xmldict(uvproj_dic['ArmAdsMisc'])
        self._uvproj_clean_xmldict(uvproj_dic['Cads'])
        self._uvproj_clean_xmldict(uvproj_dic['Cads']['VariousControls'])
        self._uvproj_clean_xmldict(uvproj_dic['LDads'])
        uvproj_dic['LDads']['ScatterFile'] = project_dic['linker_file']

        uvproj_dic['Cads']['VariousControls']['IncludePath'] = '; '.join(project_dic['include_paths'])
        uvproj_dic['Cads']['VariousControls']['Define'] = ', '.join(project_dic['macros'])
        if project_dic['macros']:
            uvproj_dic['Aads']['VariousControls']['MiscControls'] = '--cpreproc --cpreproc_opts=-D' + ',-D'.join(project_dic['macros'])

        for misc_keys in project_dic['misc'].keys():
            # ld-flags dont follow the same as asm/c flags, why?!? Please KEIL fix this
            if misc_keys == 'ld_flags':
                for item in project_dic['misc'][misc_keys]:
                    uvproj_dic[self.FLAGS_TO_UVISION[misc_keys]]['Misc'] += ' ' + item
            else:
                for item in project_dic['misc'][misc_keys]:
                    uvproj_dic[self.FLAGS_TO_UVISION[misc_keys]]['VariousControls']['MiscControls'] += ' ' + item

    def _uvproj_set_TargetCommonOption(self, uvproj_dic, project_dic):
        self._uvproj_clean_xmldict(uvproj_dic)
        self._uvproj_clean_xmldict(uvproj_dic['AfterMake'])
        self._uvproj_clean_xmldict(uvproj_dic['BeforeCompile'])
        self._uvproj_clean_xmldict(uvproj_dic['BeforeMake'])
        self._uvproj_clean_xmldict(uvproj_dic['TargetStatus'])
        uvproj_dic['OutputDirectory'] = project_dic['build_dir']
        uvproj_dic['OutputName'] = project_dic['name']
        uvproj_dic['CreateExecutable'] = 1 if project_dic['output_type'] == 'exe' else 0
        uvproj_dic['CreateLib'] = 1 if project_dic['output_type'] == 'lib' else 0

    def _uvproj_set_Utilities(self, uvproj_dic, project_dic):
        self._uvproj_clean_xmldict(uvproj_dic)

    def _uvproj_files_set(self, uvproj_dic, project_dic):
        uvproj_dic['Project']['Targets']['Target']['Groups'] = OrderedDict()
        uvproj_dic['Project']['Targets']['Target']['Groups']['Group'] = []
        i = 0
        for group_name, files in project_dic['groups'].items():
            # Why OrderedDict() - uvision project requires an order. GroupName must be before Files,
            # otherwise it does not sense any file. Same applies for other attributes, like VariousControl.
            # Therefore be aware that order matters in this exporter
            group = OrderedDict()
            group['GroupName'] = group_name
            # group['Files'] = {}
            group['Files'] = {'File': []}
            uvproj_dic['Project']['Targets']['Target']['Groups']['Group'].append(group)
            for file in files:
                uvproj_dic['Project']['Targets']['Target']['Groups']['Group'][i]['Files']['File'].append(file)
            files = uvproj_dic['Project']['Targets']['Target']['Groups']['Group'][i]['Files']['File']
            uvproj_dic['Project']['Targets']['Target']['Groups']['Group'][i]['Files']['File'] = sorted(files, key=lambda x: x['FileName'].lower())
            i += 1

    def _generate_uvmpw_file(self):
        uvmpw_dic = xmltodict.parse(open(self.uvmpw_file, "rb"))
        uvmpw_dic['ProjectWorkspace']['project'] = []

        for project in self.workspace['projects']:
            # We check how far is project from root and workspace. IF they dont match,
            # get relpath for project and inject it into workspace
            path_project = os.path.dirname(project['files']['uvproj'])
            path_workspace = os.path.dirname(self.workspace['settings']['path'] + '\\')
            destination = os.path.join(os.path.relpath(self.env_settings.root, path_project), project['files']['uvproj'])
            if path_project != path_workspace:
                destination = os.path.join(os.path.relpath(self.env_settings.root, path_workspace), project['files']['uvproj'])
            uvmpw_dic['ProjectWorkspace']['project'].append({'PathAndName': destination})

        # generate the file
        uvmpw_xml = xmltodict.unparse(uvmpw_dic, pretty=True)
        project_path, uvmpw = self.gen_file_raw(uvmpw_xml, '%s.uvmpw' % self.workspace['settings']['name'], self.workspace['settings']['path'])
        return project_path, uvmpw

    def _set_target(self, expanded_dic, uvproj_dic, tool_name):
        pro_def = ProGenDef(tool_name)
        if not pro_def.is_supported(expanded_dic['target'].lower()):
            raise RuntimeError("Target %s is not supported. Please add them to https://github.com/project-generator/project_generator_definitions" % expanded_dic['target'].lower())
        mcu_def_dic = pro_def.get_tool_definition(expanded_dic['target'].lower())
        if not mcu_def_dic:
             raise RuntimeError(
                "Target definitions were not found for %s. Please add them to https://github.com/project-generator/project_generator_definitions" % expanded_dic['target'].lower())
        logger.debug("Mcu definitions: %s" % mcu_def_dic)
        uvproj_dic['Project']['Targets']['Target']['TargetOption']['TargetCommonOption']['Device'] = mcu_def_dic['TargetOption']['Device'][0]
        uvproj_dic['Project']['Targets']['Target']['TargetOption']['TargetCommonOption']['DeviceId'] = mcu_def_dic['TargetOption']['DeviceId'][0]
        try:
            uvproj_dic['Project']['Targets']['Target']['TargetOption']['TargetCommonOption']['Vendor'] = mcu_def_dic['TargetOption']['Vendor'][0]
            uvproj_dic['Project']['Targets']['Target']['TargetOption']['TargetCommonOption']['Cpu'] = mcu_def_dic['TargetOption']['Cpu'][0]
            uvproj_dic['Project']['Targets']['Target']['TargetOption']['TargetCommonOption']['FlashDriverDll'] = str(mcu_def_dic['TargetOption']['FlashDriverDll'][0])
            uvproj_dic['Project']['Targets']['Target']['TargetOption']['TargetCommonOption']['SFDFile'] = mcu_def_dic['TargetOption']['SFDFile'][0]
            uvproj_dic['Project']['Targets']['Target']['TargetOption']['TargetCommonOption']['RegisterFile'] = mcu_def_dic['TargetOption']['RegisterFile'][0]
        except KeyError:
            pass

        # overwrite the template if target has defined debugger
        # later progen can overwrite this if debugger is set in project data
        try:
            debugger_name = pro_def.get_debugger(expanded_dic['target'])['name']
            uvproj_dic['Project']['Targets']['Target']['TargetOption']['DebugOption']['TargetDlls']['Driver'] = self.definitions.debuggers[debugger_name]['uvproj']['TargetDlls']['Driver']
            uvproj_dic['Project']['Targets']['Target']['TargetOption']['Utilities']['Flash2'] = self.definitions.debuggers[debugger_name]['uvproj']['Utilities']['Flash2']
        except (TypeError, KeyError) as err:
            pass
        # Support new device packs
        if 'PackID' in  mcu_def_dic['TargetOption']:
            if tool_name != 'uvision5':
                # using software packs require v5
                logger.info("The target might not be supported in %s, requires uvision5" % tool_name)
            uvproj_dic['Project']['Targets']['Target']['TargetOption']['TargetCommonOption']['PackID'] = mcu_def_dic['TargetOption']['PackID'][0]

    def _uvoptx_set_debugger(self, expanded_dic, uvoptx_dic, tool_name):
        pro_def = ProGenDef(tool_name)
        if not pro_def.is_supported(expanded_dic['target'].lower()):
            raise RuntimeError("Target %s is not supported. Please add them to https://github.com/project-generator/project_generator_definitions" % expanded_dic['target'].lower())
        mcu_def_dic = pro_def.get_tool_definition(expanded_dic['target'].lower())
        if not mcu_def_dic:
             raise RuntimeError(
                "Target definitions were not found for %s. Please add them to https://github.com/project-generator/project_generator_definitions" % expanded_dic['target'].lower())
        logger.debug("Mcu definitions: %s" % mcu_def_dic)

        # set the same target name FlashDriverDll config as in uvprojx file
        try:
            uvoptx_dic['ProjectOpt']['Target']['TargetName'] = expanded_dic['name']
            uvoptx_dic['ProjectOpt']['Target']['TargetOption']['TargetDriverDllRegistry']['SetRegEntry']['Name'] = str(mcu_def_dic['TargetOption']['FlashDriverDll'][0])
        except KeyError:
            return

        # load debugger from target dictionary or use default debugger
        try:
            debugger_dic = pro_def.get_debugger(expanded_dic['target'])
            if debugger_dic is None:
                debugger_name = self.definitions.debuggers_default
            else:
                debugger_name = debugger_dic['name']
            uvoptx_dic['ProjectOpt']['Target']['TargetOption']['DebugOpt']['nTsel'] = self.definitions.debuggers[debugger_name]['uvoptx']['DebugOpt']['nTsel']
            uvoptx_dic['ProjectOpt']['Target']['TargetOption']['DebugOpt']['pMon'] = self.definitions.debuggers[debugger_name]['uvoptx']['DebugOpt']['pMon']
            uvoptx_dic['ProjectOpt']['Target']['TargetOption']['TargetDriverDllRegistry']['SetRegEntry']['Key'] = self.definitions.debuggers[debugger_name]['uvoptx']['SetRegEntry']['Key']
        except KeyError:
            raise RuntimeError("Debugger %s is not supported" % expanded_dic['debugger'])

    def _export_single_project(self, tool_name):
        expanded_dic = self.workspace.copy()

        groups = self._get_groups(self.workspace)
        expanded_dic['groups'] = {}
        for group in groups:
            expanded_dic['groups'][group] = []

        # get relative path and fix all paths within a project
        self._iterate(self.workspace, expanded_dic)

        expanded_dic['build_dir'] = '.\\' + expanded_dic['build_dir'] + '\\'

        # generic tool template specified or project
        if expanded_dic['template']:
            for template in expanded_dic['template']:
                template = join(getcwd(), template)
                if os.path.splitext(template)[1] == '.uvproj' or os.path.splitext(template)[1] == '.uvprojx' or \
                    re.match('.*\.uvproj.tmpl$', template) or re.match('.*\.uvprojx.tmpl$', template):
                    try:
                        uvproj_dic = xmltodict.parse(open(template, encoding="utf8").read())
                    except IOError:
                        logger.info("Template file %s not found" % template)
                        return None, None
                else:
                    logger.info("Template file %s contains unknown template extension (.uvproj/x are valid). Using default one" % template)
                    uvproj_dic = xmltodict.parse(open(self.uvproj_file, "rb"))
        elif 'uvision' in self.env_settings.templates.keys():
            # template overrides what is set in the yaml files
            for template in self.env_settings.templates['uvision']:
                template = join(getcwd(), template)
                if os.path.splitext(template)[1] == '.uvproj' or os.path.splitext(template)[1] == '.uvprojx' or \
                    re.match('.*\.uvproj.tmpl$', template) or re.match('.*\.uvprojx.tmpl$', template):
                    try:
                        uvproj_dic = xmltodict.parse(open(template, encoding="utf8").read())
                    except IOError:
                        logger.info("Template file %s not found. Using default template" % template)
                        uvproj_dic = xmltodict.parse(open(self.uvproj_file, "rb"))
                else:
                    logger.info("Template file %s contains unknown template extension (.uvproj/x are valid). Using default one" % template)
                    uvproj_dic = xmltodict.parse(open(self.uvproj_file))
        else:
            uvproj_dic = xmltodict.parse(open(self.uvproj_file, "rb"))

        try:
            uvproj_dic['Project']['Targets']['Target']['TargetName'] = expanded_dic['name']
        except KeyError:
            raise RuntimeError("The uvision template is not valid .uvproj file")

        self._uvproj_files_set(uvproj_dic, expanded_dic)
        self._uvproj_set_CommonProperty(
            uvproj_dic['Project']['Targets']['Target']['TargetOption']['CommonProperty'], expanded_dic)
        self._uvproj_set_DebugOption(
            uvproj_dic['Project']['Targets']['Target']['TargetOption']['DebugOption'], expanded_dic)
        self._uvproj_set_DllOption(
            uvproj_dic['Project']['Targets']['Target']['TargetOption']['DllOption'], expanded_dic)
        self._uvproj_set_TargetArmAds(
            uvproj_dic['Project']['Targets']['Target']['TargetOption']['TargetArmAds'], expanded_dic)
        self._uvproj_set_TargetCommonOption(
            uvproj_dic['Project']['Targets']['Target']['TargetOption']['TargetCommonOption'], expanded_dic)
        self._uvproj_set_Utilities(
            uvproj_dic['Project']['Targets']['Target']['TargetOption']['Utilities'], expanded_dic)

        # set target only if defined, otherwise use from template/default one
        if tool_name == 'uvision5':
            extension = 'uvprojx'
            uvproj_dic['Project']['SchemaVersion'] = '2.1'
        else:
            extension = 'uvproj'
            uvproj_dic['Project']['SchemaVersion'] = '1.1'

        if expanded_dic['target']:
            self._set_target(expanded_dic, uvproj_dic, tool_name)

        # load debugger
        if expanded_dic['debugger']:
            try:
                uvproj_dic['Project']['Targets']['Target']['TargetOption']['DebugOption']['TargetDlls']['Driver'] = self.definitions.debuggers[expanded_dic['debugger']]['uvproj']['TargetDlls']['Driver']
                uvproj_dic['Project']['Targets']['Target']['TargetOption']['Utilities']['Flash2'] = self.definitions.debuggers[expanded_dic['debugger']]['uvproj']['Utilities']['Flash2']
            except KeyError:
                raise RuntimeError("Debugger %s is not supported" % expanded_dic['debugger'])

        # Project file
        uvproj_xml = xmltodict.unparse(uvproj_dic, pretty=True)
        project_path, uvproj = self.gen_file_raw(uvproj_xml, '%s.%s' % (expanded_dic['name'], extension), expanded_dic['output_dir']['path'])

        uvoptx = None

        # generic tool template specified
        uvoptx_dic = xmltodict.parse(open(self.uvoptx_file, "rb"))

        self._uvoptx_set_debugger(expanded_dic, uvoptx_dic, tool_name)

        # set target only if defined, otherwise use from template/default one
        if tool_name == 'uvision5':
            extension = 'uvoptx'
        else:
            extension = 'uvopt'

        # Project file
        uvoptx_xml = xmltodict.unparse(uvoptx_dic, pretty=True)
        project_path, uvoptx = self.gen_file_raw(uvoptx_xml, '%s.%s' % (expanded_dic['name'], extension), expanded_dic['output_dir']['path'])

        return project_path, [uvproj, uvoptx]

    def export_workspace(self):
        path, workspace = self._generate_uvmpw_file()
        return path, [workspace]

    def export_project(self):
        path, files = self._export_single_project('uvision') #todo: uvision will switch to uv4
        generated_projects = copy.deepcopy(self.generated_project)
        generated_projects['path'] = path
        generated_projects['files']['uvproj'] = files[0]
        return generated_projects

    def get_generated_project_files(self):
        return {'path': self.workspace['path'], 'files': [self.workspace['files']['uvproj']]}

    def _build_project(self, tool_name, extension):
        # > UV4 -b [project_path]
        path = join(self.env_settings.root, self.workspace['files'][extension])
        if path.split('.')[-1] != extension:
            path = path + extension

        if not os.path.exists(path):
            logger.debug("The file: %s does not exists, exported prior building?" % path)
            return -1

        logger.debug("Building uVision project: %s" % path)

        build_log_path = join(os.path.dirname(path),'build','build_log.txt')
        args = [self.env_settings.get_env_settings(tool_name), '-r', '-j0', '-o', build_log_path, path]
        logger.debug(args)

        try:
            ret_code = None
            ret_code = subprocess.call(args)
        except:
            logger.error(
                "Error whilst calling UV4: '%s'. Please set uvision path in the projects.yaml file." % self.env_settings.get_env_settings('uvision'))
            return -1
        else:
            if ret_code != self.SUCCESSVALUE and ret_code != self.WARNVALUE:
                # Seems like something went wrong.
                logger.error("Project: %s build failed with the status: %s" % (self.workspace['files'][extension], self.ERRORLEVEL.get(ret_code, "Unknown")))
                return -1
            else:
                logger.info("Project: %s build succeeded with the status: %s" % (self.workspace['files'][extension], self.ERRORLEVEL.get(ret_code, "Unknown")))
                return 0

    def build_project(self, **kwargs):
        return self._build_project('uvision', 'uvproj')

class Uvision5(Uvision):

    generated_project = {
        'path': '',
        'files': {
            'uvprojx': '',
            'uvoptx': '',
        }
    }

    def __init__(self, workspace, env_settings):
        super(Uvision5, self).__init__(workspace, env_settings)

    @staticmethod
    def get_toolnames():
        return ['uvision5']

    def export_project(self):
        path, files = self._export_single_project('uvision5')
        generated_projects = copy.deepcopy(self.generated_project)
        generated_projects['path'] = path
        generated_projects['files']['uvprojx'] = files[0]
        generated_projects['files']['uvoptx'] = files[1]
        return generated_projects

    def get_generated_project_files(self):
        return {'path': self.workspace['path'], 'files': [self.workspace['files']['uvprojx'], self.workspace['files']['uvoptx']]}

    def build_project(self, **kwargs):
        # tool_name uvision as uv4 is still used in uv5
        return self._build_project('uvision', 'uvprojx')



# Copyright 2015 0xc0170
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import os
import yaml
import shutil

from unittest import TestCase

from project_generator.commands import generate, clean
from .simple_project import project_1_yaml, projects_yaml, project_2_yaml

class TestCleanCommand(TestCase):

    """test clean command"""

    def setUp(self):
        if not os.path.exists('test_workspace'):
            os.makedirs('test_workspace')
        # write project file
        with open(os.path.join(os.getcwd(), 'test_workspace/project_1.yaml'), 'wt') as f:
            f.write(yaml.dump(project_1_yaml, default_flow_style=False))
        # write project file
        with open(os.path.join(os.getcwd(), 'test_workspace/project_2.yaml'), 'wt') as f:
            f.write(yaml.dump(project_2_yaml, default_flow_style=False))
        # write projects file
        with open(os.path.join(os.getcwd(), 'test_workspace/projects.yaml'), 'wt') as f:
            f.write(yaml.dump(projects_yaml, default_flow_style=False))

        self.parser = argparse.ArgumentParser()
        subparsers = self.parser.add_subparsers(help='commands')
        self.export_subparser = subparsers.add_parser('generate', help=generate.help)
        self.clean_subparser = subparsers.add_parser('clean', help=clean.help)

    def tearDown(self):
        # remove created directory
        shutil.rmtree('test_workspace', ignore_errors=True)
        shutil.rmtree('generated_projects', ignore_errors=True)


    def test_clean_one_project(self):
        # We first generate project, then clean it
        generate.setup(self.export_subparser)
        args = self.parser.parse_args(['generate','-f','test_workspace/projects.yaml','-p','project_2',
            '-t', 'uvision'])
        result = generate.run(args)

        # this should generate a project to generated_projects/uvision_project_2/project_2.uvproj
        assert os.path.isfile('generated_projects/uvision_project_2/project_2.uvproj')
        assert result == 0

        # now clean
        clean.setup(self.clean_subparser)
        args = self.parser.parse_args(['clean','-f','test_workspace/projects.yaml','-p','project_2',
            '-t', 'uvision'])
        result = clean.run(args)

        assert not os.path.isfile('generated_projects/uvision_project_2/project_2.uvproj')
        assert not os.path.isdir('generated_projects/uvision_project_2')
        assert result == 0



# Copyright 2014-2015 0xc0170
# Copyright (c) 2020 Chris Reed
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .tools.iar import IAREmbeddedWorkbench
from .tools.uvision import Uvision, Uvision5
from .tools.coide import Coide
from .tools.eclipse import EclipseGnuARM
from .tools.gccarm import MakefileGccArm
from .tools.makearmcc import MakefileArmcc
from .tools.makearmclang import MakefileArmclang
from .tools.sublimetext import SublimeTextMakeGccARM
from .tools.gdb import GDB, ARMNoneEABIGDB, JLinkGDB
from .tools.cmakegccarm import CMakeGccArm
from .tools.cmakearmcc import CMakeArmcc
from .tools.cmakearmclang import CMakeArmClang
from .tools.visual_studio import VisualStudioMakeGCCARM, VisualStudioGDB

class ToolsSupported:
    """ Represents all tools available """

    # Default tools - aliases
    TOOLS_ALIAS = {
        'uvision':       'uvision4',
        'iar':           'iar_arm',
        'make_gcc':      'make_gcc_arm',
        'gcc_arm':       'make_gcc_arm',
        'armcc':         'make_armcc',
        'armclang':      'make_armclang',
        'eclipse':       'eclipse_make_gcc_arm',
        'sublime':       'sublime_make_gcc_arm',
        'sublime_text':  'sublime_make_gcc_arm',
        'visual_studio': 'visual_studio_make_gcc_arm',
        'cmake':         'cmake_gcc_arm',
     }

    # Tools dictionary
    # Each of this tool needs to support at least:
    # - get_toolchain (toolchain is a list of toolchains supported by tool)
    # - get_toolname (returns name string)
    # - export_project (basic functionality to be covered by a tool)
    TOOLS_DICT = {
        'iar_arm':              IAREmbeddedWorkbench,
        'uvision4':             Uvision,
        'uvision5':             Uvision5,
        'coide':                Coide,
        'make_gcc_arm':         MakefileGccArm,
        'make_armcc':           MakefileArmcc, 
        'make_armclang':        MakefileArmclang,
        'eclipse_make_gcc_arm': EclipseGnuARM,
        'sublime_make_gcc_arm': SublimeTextMakeGccARM,
        'gdb':                  GDB,
        'arm_none_eabi_gdb':    ARMNoneEABIGDB,
        'jlink_gdb':            JLinkGDB,
        'cmake_gcc_arm':        CMakeGccArm,
        'cmake_armcc':          CMakeArmcc,
        'cmake_armclang':       CMakeArmClang,
        'visual_studio_gdb':    VisualStudioGDB,
        'visual_studio_make_gcc_arm': VisualStudioMakeGCCARM,
    }

    TOOLCHAINS = list(set([v.get_toolchain() for k, v in TOOLS_DICT.items() if v.get_toolchain() is not None]))
    TOOLS = list(set([v for k, v in TOOLS_DICT.items() if v is not None]))

    def _get_tool_name(self, tool):
        if tool in self.TOOLS_ALIAS.keys():
            tool = self.TOOLS_ALIAS[tool]
        return tool

    def get_tool(self, tool):
        name = self._get_tool_name(tool)
        try:
            return self.TOOLS_DICT[name]
        except KeyError:
            return None

    def get_toolnames(self, tool):
        name = self._get_tool_name(tool)
        try:
            return self.TOOLS_DICT[name].get_toolnames()
        except KeyError:
            return None

    def get_toolchain(self, tool):
        name = self._get_tool_name(tool)
        try:
            return self.TOOLS_DICT[name].get_toolchain()
        except KeyError:
            return None

    def get_supported(self):
        return list(self.TOOLS_DICT.keys()) + list(self.TOOLS_ALIAS.keys()) + self.TOOLCHAINS



# Copyright 2015 0xc0170
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import os
import yaml
import shutil

from unittest import TestCase

from project_generator.commands import list_projects
from .simple_project import project_1_yaml, projects_yaml, project_2_yaml

class TestListCommand(TestCase):

    """test list command"""

    def setUp(self):
        self.parser = argparse.ArgumentParser()
        subparsers = self.parser.add_subparsers(help='commands')
        self.subparser = subparsers.add_parser('list', help=list_projects.help)

    def tearDown(self):
        # remove created directory
        shutil.rmtree('test_workspace', ignore_errors=True)

    def test_list_projects(self):
        # For now list projects should not fail, we should add checking stdout if projects match
        list_projects.setup(self.subparser)
        args = self.parser.parse_args(['list','projects','--all'])
        result = list_projects.run(args)

        assert result == 0

    def test_list_targets(self):
        # For now list projects should not fail, we should add checking stdout if targets match
        list_projects.setup(self.subparser)
        args = self.parser.parse_args(['list','targets','--all'])
        result = list_projects.run(args)

        assert result == 0

    def test_list_tools(self):
        # For now list projects should not fail, we should add checking stdout if tools match
        list_projects.setup(self.subparser)
        args = self.parser.parse_args(['list','tools','--all'])
        result = list_projects.run(args)

        assert result == 0



# Copyright 2014-2015 0xc0170
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from os.path import normpath, join, pardir, sep

class ProjectSettings:
    PROJECT_ROOT = os.environ.get('PROJECT_GENERATOR_ROOT') or join(pardir, pardir)
    DEFAULT_TOOL = os.environ.get('PROJECT_GENERATOR_DEFAULT_TOOL') or 'uvision'

    DEFAULT_EXPORT_LOCATION_FORMAT = join('generated_projects', '{tool}_{project_name}')
    DEFAULT_ROOT = os.getcwd()

    def __init__(self):
        """ This are default enviroment settings for build tools. To override,
        define them in the projects.yaml file. """

        self.paths = {}
        self.templates = {}
        self.paths['uvision'] = os.environ.get('UV4') or join('C:', sep,
            'Keil', 'UV4', 'UV4.exe')
        self.paths['iar'] = os.environ.get('IARBUILD') or join(
            'C:', sep, 'Program Files (x86)',
            'IAR Systems', 'Embedded Workbench 7.0',
            'common', 'bin')
        self.paths['gcc'] = os.environ.get('ARM_GCC_PATH') or ''

        self.export_location_format = self.DEFAULT_EXPORT_LOCATION_FORMAT
        self.root = os.getcwd()

    def update(self, settings):
        if settings:
            if 'tools' in settings:
                for k, v in settings['tools'].items():
                    if k in self.paths:
                        if 'path' in v.keys():
                            self.paths[k] = v['path'][0]
                    if 'template' in v.keys():
                        self.templates[k] = v['template']

            if 'export_dir' in settings:
                self.export_location_format = normpath(settings['export_dir'][0])
            if 'root' in settings:
                self.root = normpath(settings['root'][0])

    def get_env_settings(self, env_set):
        return self.paths[env_set]



# Copyright 2015 0xc0170
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from project_generator.util import *

def test_flatten():
    l1 = [['aa', 'bb', ['cc', 'dd', 'ee'], ['ee', 'ff'], 'gg']]

    assert list(flatten(l1)) == ['aa', 'bb', 'cc', 'dd', 'ee', 'ee', 'ff', 'gg']
    assert uniqify(flatten(l1)) == ['aa', 'bb', 'cc', 'dd', 'ee', 'ff', 'gg']

def test_uniqify():
    l1 = ['a', 'b', 'b', 'c', 'b', 'd', 'c', 'e', 'f', 'a']
    assert uniqify(l1) == ['a', 'b', 'c', 'd', 'e', 'f']



# Copyright 2015 0xc0170
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import os
import yaml
import shutil

from unittest import TestCase

from project_generator.commands import generate
from .simple_project import project_1_yaml, projects_yaml, project_2_yaml

class TestExportCommand(TestCase):

    """test generate command"""

    def setUp(self):
        if not os.path.exists('test_workspace'):
            os.makedirs('test_workspace')
        # write project file
        with open(os.path.join(os.getcwd(), 'test_workspace/project_1.yaml'), 'wt') as f:
            f.write(yaml.dump(project_1_yaml, default_flow_style=False))
        # write project file
        with open(os.path.join(os.getcwd(), 'test_workspace/project_2.yaml'), 'wt') as f:
            f.write(yaml.dump(project_2_yaml, default_flow_style=False))
        # write projects file
        with open(os.path.join(os.getcwd(), 'test_workspace/projects.yaml'), 'wt') as f:
            f.write(yaml.dump(projects_yaml, default_flow_style=False))

        self.parser = argparse.ArgumentParser()
        subparsers = self.parser.add_subparsers(help='commands')
        self.subparser = subparsers.add_parser('generate', help=generate.help)

    def tearDown(self):
        # remove created directory
        shutil.rmtree('test_workspace', ignore_errors=True)
        shutil.rmtree('generated_projects', ignore_errors=True)

    def test_generate_project3_all_tools(self):
        generate.setup(self.subparser)
        args = self.parser.parse_args(['generate','-f','test_workspace/projects.yaml','-p','project_3'])
        result = generate.run(args)

        # one of the tools is unknown , should return -1
        assert result == -1

    def test_generate_project2_all_tools(self):
        generate.setup(self.subparser)
        args = self.parser.parse_args(['generate','-f','test_workspace/projects.yaml','-p','project_2'])
        result = generate.run(args)

        # No tools defined
        assert result == -1

    def test_generate_one_project_uvision(self):
        generate.setup(self.subparser)
        args = self.parser.parse_args(['generate','-f','test_workspace/projects.yaml','-p','project_2',
            '-t', 'uvision'])
        result = generate.run(args)

        assert result == 0

        # this should generate a project to generated_projects/uvision_project_2/project_2.uvproj
        assert os.path.isfile('generated_projects/uvision_project_2/project_2.uvproj')

    def test_generate_one_project_iar_arm(self):
        generate.setup(self.subparser)
        args = self.parser.parse_args(['generate','-f','test_workspace/projects.yaml','-p','project_2',
            '-t', 'iar_arm'])
        result = generate.run(args)

        assert result == 0

        # this should generate a project to generated_projects/uvision_project_2/project_2.ewp/ewd/eww
        assert os.path.isfile('generated_projects/iar_arm_project_2/project_2.ewp')
        assert os.path.isfile('generated_projects/iar_arm_project_2/project_2.ewd')
        assert os.path.isfile('generated_projects/iar_arm_project_2/project_2.eww')

    def test_generate_one_project_make_gcc_arm(self):
        generate.setup(self.subparser)
        args = self.parser.parse_args(['generate','-f','test_workspace/projects.yaml','-p','project_2',
            '-t', 'make_gcc_arm'])
        result = generate.run(args)

        assert result == 0
        # this should generate a project to generated_projects/uvision_project_2/Makefile
        assert os.path.isfile('generated_projects/make_gcc_arm_project_2/Makefile')

    def test_generate_one_project_coide(self):
        generate.setup(self.subparser)
        args = self.parser.parse_args(['generate','-f','test_workspace/projects.yaml','-p','project_2',
            '-t', 'coide'])
        result = generate.run(args)

        assert result == 0

        # this should generate a project to generated_projects/uvision_project_2/project_2.coproj
        assert os.path.isfile('generated_projects/coide_project_2/project_2.coproj')

    def test_generate_one_project_arm_none_eabi_gdb(self):
        generate.setup(self.subparser)
        args = self.parser.parse_args(['generate','-f','test_workspace/projects.yaml','-p','project_2',
            '-t', 'arm_none_eabi_gdb'])
        result = generate.run(args)

        assert result == 0

    def test_generate_one_project_sublime_make_gcc_arm(self):
        generate.setup(self.subparser)
        args = self.parser.parse_args(['generate','-f','test_workspace/projects.yaml','-p','project_2',
            '-t', 'sublime_make_gcc_arm'])
        result = generate.run(args)

        assert result == 0

    def test_generate_one_project_eclipse_make_gcc_arm(self):
        generate.setup(self.subparser)
        args = self.parser.parse_args(['generate','-f','test_workspace/projects.yaml','-p','project_2',
            '-t', 'eclipse_make_gcc_arm'])
        result = generate.run(args)

        assert result == 0

    def test_generate_workspace_all_tools(self):
        generate.setup(self.subparser)
        args = self.parser.parse_args(['generate','-f','test_workspace/projects.yaml','-p',
            'project_workspace'])
        result = generate.run(args)

        # we dont specify tool to generate, which is not valid for workspace.
        # we don't know which tool we should build worksapce for as it consists
        # of projects, and each can speficify tools supported.
        assert result == -1

    def test_generate_workspace_uvision(self):
        generate.setup(self.subparser)
        args = self.parser.parse_args(['generate','-f','test_workspace/projects.yaml','-p',
            'project_workspace', '-t', 'uvision'])
        result = generate.run(args)

        assert result == 0

    def test_generate_workspace_iar_arm(self):
        generate.setup(self.subparser)
        args = self.parser.parse_args(['generate','-f','test_workspace/projects.yaml','-p',
            'project_workspace', '-t', 'iar_arm'])
        result = generate.run(args)

        assert result == 0



# Copyright 2014-2015 0xc0170
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import logging

from ..tools_supported import ToolsSupported
from ..generate import Generator
from . import argparse_filestring_type, argparse_string_type

help = 'Clean generated projects'


def run(args):
    combined_projects = args.projects + args.project or ['']
    generator = Generator(args.file)
    for project_name in (combined_projects):
        for project in generator.generate(project_name):
            project.clean(args.tool)
    return 0

def setup(subparser):
    subparser.add_argument('-v', dest='verbosity', action='count', default=0,
                        help='Increase the verbosity of the output (repeat for more verbose output)')
    subparser.add_argument('-q', dest='quietness', action='count', default=0,
                        help='Decrease the verbosity of the output (repeat for more verbose output)')
    subparser.add_argument("-f", "--file", help="YAML projects file", default='projects.yaml', type=argparse_filestring_type)
    subparser.add_argument("-p", "--project", dest="projects", action='append', default=[],
                        help="Specify which project to be removed")
    subparser.add_argument(
        "-t", "--tool", help="Clean project files for this tool", required=True,
        type=argparse_string_type(str.lower, False), choices=list(ToolsSupported.TOOLS_DICT.keys()) + list(ToolsSupported.TOOLS_ALIAS.keys()))
    subparser.add_argument("project", nargs='*',
                        help="Specify projects to be removed")



# Copyright 2014-2015 0xc0170
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import yaml

from .settings import ProjectSettings
from .util import flatten, uniqify, load_yaml_records
from .project import *

class Generator:
    def __init__(self, projects_file):
        if type(projects_file) is not dict:
            try:
                with open(projects_file, 'rt') as f:
                    self.projects_dict = yaml.load(f, Loader=yaml.FullLoader)
            except IOError:
               raise IOError("The main progen projects file %s doesn't exist." % projects_file)
        else:
            self.projects_dict = projects_file
        self.workspaces = {}
        self.settings = ProjectSettings()

        if 'settings' in self.projects_dict:
            self.settings.update(self.projects_dict['settings'])

    def generate(self, name=''):
        found = False
        if name != '':
            if (name == 'global_workspace'):
                if 'workspaces' not in self.projects_dict:
                    self.projects_dict['workspaces'] = {}
                if  ('global_workspace' not in self.projects_dict['workspaces']):
                    self.projects_dict['workspaces']['global_workspace'] = {
                        "projects": list(self.projects_dict['projects'].keys())
                    }
            # process project first, workspaces afterwards
            if 'projects' in self.projects_dict:
                if name in self.projects_dict['projects'].keys():
                    found = True
                    records = self.projects_dict['projects'][name]
                    yield Project(name, load_yaml_records(uniqify(flatten(records))), self.settings)
            if 'workspaces' in self.projects_dict:
                workspace_settings = {}
                if name in self.projects_dict['workspaces'].keys():
                    found = True
                    records = self.projects_dict['workspaces'][name]
                    if 'settings' in records:
                        workspace_settings = records['settings'] 
                    workspace_settings['name'] = name
                    projects = [Project(project, load_yaml_records(uniqify(flatten(self.projects_dict['projects'][project]))), self.settings, name) for project in records['projects']]
                    self.workspaces[name] = ProjectWorkspace(name, projects, self.settings, workspace_settings)
                    yield self.workspaces[name]
        else:
            if 'projects' in self.projects_dict:
                found = True
                for name, records in sorted(self.projects_dict['projects'].items(),
                                            key=lambda x: x[0]):
                    yield Project(name, load_yaml_records(uniqify(flatten(records))), self.settings)
            if 'workspaces' in self.projects_dict:
                found = True
                for name, records in sorted(self.projects_dict['workspaces'].items(),
                                            key=lambda x: x[0]):
                    workspace_settings = {}
                    if 'settings' in records:
                        workspace_settings = records['settings']
                    workspace_settings['name'] = name
                    projects = [Project(project, load_yaml_records(uniqify(flatten(self.projects_dict['projects'][project]))), self.settings, name) for project in records['projects']]
                    self.workspaces[name] = ProjectWorkspace(name, projects, self.settings, workspace_settings)
                    yield self.workspaces[name]

        if not found:
            logging.error("You specified an invalid project name.")



project_1_yaml = {
    'common': {
        'sources': {'sources': ['test_workspace/main.cpp']},
        'includes': {'includes': ['test_workspace/header1.h']},
        'macros': ['MACRO1', 'MACRO2'],
        'target': ['mbed-lpc1768'],
        'core': ['core1'],
        'tools_supported': ['iar_arm', 'uvision', 'coide', 'unknown'],
        'output_type': ['exe'],
        'debugger': ['j-link'],
        'linker_file': ['test_workspace/linker.ld'],
    }
}

project_2_yaml = {
    'common': {
        'sources': ['test_workspace/main.cpp'],
        'includes': ['test_workspace/header1.h'],
        'macros': ['MACRO1', 'MACRO2'],
        'target': ['mbed-lpc1768'],
        'core': ['core2'],
        'output_type': ['exe'],
        'debugger': ['j-link'],
        'linker_file': ['test_workspace/linker.ld'],
    }
}

projects_yaml = {
    'projects': {
        'project_2': ['test_workspace/project_2.yaml'],
        'project_3': ['test_workspace/project_1.yaml'],
    },
    'workspaces': {
        'project_workspace': {
            'projects': ['project_3'],
        },
    }

}



# Copyright 2015 0xc0170
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import yaml
import shutil

from unittest import TestCase

from project_generator.generate import Generator
from project_generator.project import Project
from project_generator.settings import ProjectSettings
from project_generator.tools.uvision import uVisionDefinitions, Uvision5

from .simple_project import project_1_yaml, project_2_yaml, projects_1_yaml

class TestProject(TestCase):

    """test things related to the uvision tool"""

    def setUp(self):
        if not os.path.exists('test_workspace'):
            os.makedirs('test_workspace')
        # write project file
        with open(os.path.join(os.getcwd(), 'test_workspace/project_1.yaml'), 'wt') as f:
            f.write(yaml.dump(project_1_yaml, default_flow_style=False))
        with open(os.path.join(os.getcwd(), 'test_workspace/project_2.yaml'), 'wt') as f:
            f.write(yaml.dump(project_2_yaml, default_flow_style=False))
        # write projects file
        with open(os.path.join(os.getcwd(), 'test_workspace/projects.yaml'), 'wt') as f:
            f.write(yaml.dump(projects_1_yaml, default_flow_style=False))

        self.project = next(Generator(projects_1_yaml).generate('project_1'))
        self.project2 = next(Generator(projects_1_yaml).generate('project_2'))

        self.defintions = uVisionDefinitions()
        self.uvision = Uvision5(self.project.project, ProjectSettings())

    def tearDown(self):
        # remove created directory
        shutil.rmtree('test_workspace', ignore_errors=True)
        shutil.rmtree('generated_projects', ignore_errors=True)

    # this is now commented, a project needs to be adjusted before exporting, so this one
    # fails. I'll keep it for a while as a reminder
    # def test_export(self):
    #     self.uvision.export_project()

    def test_export_project(self):
        result = self.project.generate('uvision5', False)
        # it should get generated files from the last export
        projectfiles = self.project.get_generated_project_files('uvision5')

        assert result == 0
        assert projectfiles
        assert os.path.splitext(projectfiles['files'][0])[1] == '.uvprojx'

    def test_export_project_to_diff_directory(self):
        project_1_yaml['common']['export_dir'] = ['create_this_folder']
        with open(os.path.join(os.getcwd(), 'test_workspace/project_1.yaml'), 'wt') as f:
            f.write(yaml.dump(project_1_yaml, default_flow_style=False))
        for project in Generator(projects_1_yaml).generate('project_1'):
            result = project.generate('uvision5', False)

        assert result == 0
        assert os.path.isdir('create_this_folder')
        shutil.rmtree('create_this_folder')

    def test_build_project(self):
        result_export = self.project.generate('uvision5', False)
        result_build = self.project.build('uvision5')

        assert result_export == 0
        # nonvalid project, should fail with errors
        assert result_build == -1

    def test_template(self):
        # should fail as template does not exists
        result = self.project2.generate('uvision5', False)
        assert result == 0



# Copyright 2014-2015 0xc0170
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import ntpath

from copy import deepcopy
import logging

from .makefile import MakefileTool

logger = logging.getLogger('progen.tools.armcc')

class MakefileArmcc(MakefileTool):

    def __init__(self, workspace, env_settings):
        MakefileTool.__init__(self, workspace, env_settings, logger)

    @staticmethod
    def get_toolnames():
        return ['make_armcc']

    @staticmethod
    def get_toolchain():
        return 'armcc'

    def _get_libs(self, project_data):
        project_data['lib_paths'] =[]
        project_data['libraries'] =[]
        for lib in project_data['source_files_lib']:
            head, tail = ntpath.split(lib)
            file = tail
            if (os.path.splitext(file)[1] != ".lib"):
                self.logging.debug("Found %s lib with non-valid extension (!=.lib)" % file)
                continue
            else:
                lib_path = ntpath.join(self.workspace['output_dir']['path'], head)
                lib_path = ntpath.abspath(lib_path)
                project_data['lib_paths'].append(lib_path)
                project_data['libraries'].append(file)

    def export_project(self):
        """ Processes misc options specific for ARMCC, and run generator """
        generated_projects = deepcopy(self.generated_projects)
        self.process_data_for_makefile(self.workspace)
        generated_projects['path'], generated_projects['files']['makefile'] = \
            self.gen_file_jinja('makefile_armcc.tmpl', self.workspace, 'Makefile',
                                self.workspace['output_dir']['path'])
        return generated_projects



# Copyright 2015 0xc0170
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import shutil

import yaml
from unittest import TestCase

from project_generator.project import Project
from project_generator.generate import Generator
from project_generator.settings import ProjectSettings
from project_generator.util import merge_recursive

project_1_yaml = {
    'common': {
        'sources': { 'sources_dict' : ['test_workspace/main.cpp']
        },
        'includes': ['test_workspace/header1.h'],
        'macros': ['MACRO1', 'MACRO2', None],
        'target': ['target1'],
        'core': ['core1'],
        'tools_supported': ['iar_arm', 'uvision', 'coide', 'unknown'],
        'output_type': ['exe'],
        'debugger': ['debugger_1'],
        'linker_file': ['test_workspace/linker.ld'],
    }
}

project_2_yaml = {
    'common': {
        'sources': { 
            'sources_dict' : ['test_workspace/file2.cpp'],
            'sources_dict2' : ['test_workspace/file3.cpp']
        },
        'includes':  {
            'include_dict' : ['test_workspace/header2.h'],
            'include_dict2' : ['test_workspace/header3.h']
        },
        'macros': ['MACRO2_1', 'MACRO2_2'],
    }
}

projects_yaml = {
    'projects': {
        'project_1' : ['test_workspace/project_1.yaml', 'test_workspace/project_2.yaml']
    },
    'settings' : {
        'export_dir': ['projects/{tool}_{target}/{project_name}']
    }
}

def test_output_directory_formatting():
    path, depth = Project._generate_output_dir(ProjectSettings(),'aaa/bbb/cccc/ddd/eee/ffff/ggg')

    assert depth == 7
    assert os.path.normpath(path) == os.path.normpath('../../../../../../../')

class TestProjectYAML(TestCase):

    """test things related to the Project class"""

    def setUp(self):
        if not os.path.exists('test_workspace'):
            os.makedirs('test_workspace')
        # write project file
        with open(os.path.join(os.getcwd(), 'test_workspace/project_1.yaml'), 'wt') as f:
            f.write(yaml.dump(project_1_yaml, default_flow_style=False))
        with open(os.path.join(os.getcwd(), 'test_workspace/project_2.yaml'), 'wt') as f:
            f.write(yaml.dump(project_2_yaml, default_flow_style=False))
        # write projects file
        with open(os.path.join(os.getcwd(), 'test_workspace/projects.yaml'), 'wt') as f:
            f.write(yaml.dump(projects_yaml, default_flow_style=False))

        # now that Project and PgenWorkspace accepts dictionaries, we dont need to
        # create yaml files!
        self.project = next(Generator(projects_yaml).generate('project_1'))

        # create 3 files to test project
        with open(os.path.join(os.getcwd(), 'test_workspace/main.cpp'), 'wt') as f:
            pass
        with open(os.path.join(os.getcwd(), 'test_workspace/file2.cpp'), 'wt') as f:
            pass
        with open(os.path.join(os.getcwd(), 'test_workspace/file3.cpp'), 'wt') as f:
            pass
        with open(os.path.join(os.getcwd(), 'test_workspace/header1.h'), 'wt') as f:
            pass
        with open(os.path.join(os.getcwd(), 'test_workspace/header2.h'), 'wt') as f:
            pass
        with open(os.path.join(os.getcwd(), 'test_workspace/header3.h'), 'wt') as f:
            pass
        with open(os.path.join(os.getcwd(), 'test_workspace/linker.ld'), 'wt') as f:
            pass

    def tearDown(self):
        # remove created directory
        shutil.rmtree('test_workspace', ignore_errors=True)
        shutil.rmtree('projects', ignore_errors=True)

    def test_project_yaml(self):
        # test using yaml files and compare basic data
        project = next(Generator('test_workspace/projects.yaml').generate('project_1'))
        assert self.project.name == project.name
        # fix this one, they should be equal
        #self.assertDictEqual(self.project.project, project.project)

    def test_name(self):
        assert self.project.name == 'project_1'

    def test_project_attributes(self):
        self.project._fill_export_dict('uvision')
        assert set(self.project.project['export']['macros'] + [None]) & set(project_1_yaml['common']['macros'] + project_2_yaml['common']['macros']) 
        assert set(self.project.project['export']['include_files'].keys()) & set(['default'] + list(project_2_yaml['common']['includes'].keys()))

        # no c or asm files, empty dics
        assert self.project.project['export']['source_files_c'] == dict()
        assert self.project.project['export']['source_files_s'] == dict()
        # source groups should be equal
        assert self.project.project['export']['source_files_cpp'].keys() == merge_recursive(project_1_yaml['common']['sources'], project_2_yaml['common']['sources']).keys()

    def test_copy(self):
        # test copy method which should copy all files to generated project dir by default
        self.project._fill_export_dict('uvision', True)
        self.project._copy_sources_to_generated_destination()

    def test_set_output_dir_path(self):
        self.project._fill_export_dict('uvision')
        assert self.project.project['export']['output_dir']['path'] == os.path.join('projects', 'uvision_target1','project_1')

class TestProjectDict(TestCase):

    """test things related to the Project class, using python dicts"""

    def setUp(self):
        if not os.path.exists('test_workspace'):
            os.makedirs('test_workspace')

        # create 3 files to test project
        with open(os.path.join(os.getcwd(), 'test_workspace/main.cpp'), 'wt') as f:
            pass
        with open(os.path.join(os.getcwd(), 'test_workspace/file2.cpp'), 'wt') as f:
            pass
        with open(os.path.join(os.getcwd(), 'test_workspace/file3.cpp'), 'wt') as f:
            pass
        with open(os.path.join(os.getcwd(), 'test_workspace/header1.h'), 'wt') as f:
            pass
        with open(os.path.join(os.getcwd(), 'test_workspace/header2.h'), 'wt') as f:
            pass
        with open(os.path.join(os.getcwd(), 'test_workspace/header3.h'), 'wt') as f:
            pass
        with open(os.path.join(os.getcwd(), 'test_workspace/linker.ld'), 'wt') as f:
            pass

        self.project = Project('project_1', [project_1_yaml, project_2_yaml], ProjectSettings())

    def tearDown(self):
        # remove created directory
        shutil.rmtree('test_workspace', ignore_errors=True)

    def test_name(self):
        assert self.project.name == 'project_1'

    def test_project_attributes(self):
        self.project._fill_export_dict('uvision')
        assert set(self.project.project['export']['macros'] + [None]) & set(project_1_yaml['common']['macros'] + project_2_yaml['common']['macros']) 
        assert set(self.project.project['export']['include_files'].keys()) & set(['default'] + list(project_2_yaml['common']['includes'].keys()))

        # no c or asm files, empty dics
        assert self.project.project['export']['source_files_c'] == dict()
        assert self.project.project['export']['source_files_s'] == dict()
        # source groups should be equal
        assert self.project.project['export']['source_files_cpp'].keys() == merge_recursive(project_1_yaml['common']['sources'], project_2_yaml['common']['sources']).keys()

    def test_sources_groups(self):
        self.project._fill_export_dict('uvision')
        source_groups = []
        for dic in self.project.project['export']['sources']:
            for k, v in dic.items():
                source_groups.append(k)
        assert source_groups == list(project_1_yaml['common']['sources'].keys()) + list(project_2_yaml['common']['sources'].keys())

    def test_copy(self):
        # test copy method which should copy all files to generated project dir by default
        self.project._fill_export_dict('uvision', True)
        self.project._copy_sources_to_generated_destination()

    def test_set_output_dir_path(self):
        self.project._fill_export_dict('uvision')
        # we use default one in this class
        assert self.project.project['export']['output_dir']['path'] == os.path.join('generated_projects', 'uvision_project_1')



# Copyright 2014-2015 0xc0170
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import logging

from collections import OrderedDict
# eclipse works with linux paths
from posixpath import normpath, join, basename

from .tool import Tool, Builder, Exporter
from .gccarm import MakefileGccArm
from ..util import SOURCE_KEYS

logger = logging.getLogger('progen.tools.eclipse')

class EclipseGnuARM(Tool, Exporter, Builder):

    file_types = {'cpp': 1, 'c': 1, 's': 1, 'obj': 1, 'lib': 1, 'h': 1}

    generated_project = {
        'path': '',
        'files': {
            'proj_file': '',
            'cproj': '',
            'makefile': '',
        }
    }

    def __init__(self, workspace, env_settings):
        self.definitions = 0
        self.exporter = MakefileGccArm(workspace, env_settings)
        self.workspace = workspace
        self.env_settings = env_settings

    @staticmethod
    def get_toolnames():
        return ['eclipse_make_gcc_arm', 'make_gcc_arm']

    @staticmethod
    def get_toolchain():
        return 'gcc_arm'

    def _expand_one_file(self, source, new_data, extension):
        # Remove rel prefix.
        rel_path = new_data['output_dir']['rel_path']
        if source.startswith(rel_path):
            non_rel_source = source[len(rel_path):]
        return {"path": join('PARENT-%s-PROJECT_LOC' % new_data['output_dir']['rel_count'], normpath(non_rel_source)), "name": basename(
                    non_rel_source), "type": self.file_types[extension.lower()]}

    def _expand_sort_key(self, file) :
        return file['name'].lower()

    def export_workspace(self):
        logger.debug("Current version of CoIDE does not support workspaces")

    def export_project(self):
        """ Processes groups and misc options specific for eclipse, and run generator """

        output = copy.deepcopy(self.generated_project)
        data_for_make = self.workspace.copy()

        self.exporter.process_data_for_makefile(data_for_make)
        output['path'], output['files']['makefile'] = self.gen_file_jinja('makefile_gcc.tmpl', data_for_make, 'Makefile', data_for_make['output_dir']['path'])

        expanded_dic = self.workspace.copy()
        expanded_dic['rel_path'] = data_for_make['output_dir']['rel_path']
        groups = self._get_groups(expanded_dic)
        expanded_dic['groups'] = {}
        for group in groups:
            expanded_dic['groups'][group] = []
        self._iterate(self.workspace, expanded_dic)

        # Project file
        project_path, output['files']['cproj'] = self.gen_file_jinja(
            'eclipse_makefile.cproject.tmpl', expanded_dic, '.cproject', data_for_make['output_dir']['path'])
        project_path, output['files']['proj_file'] = self.gen_file_jinja(
            'eclipse.project.tmpl', expanded_dic, '.project', data_for_make['output_dir']['path'])
        return output

    def get_generated_project_files(self):
        return {'path': self.workspace['path'], 'files': [self.workspace['files']['proj_file'], self.workspace['files']['cproj'],
            self.workspace['files']['makefile']]}




# Copyright 2015 0xc0170
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
from os.path import exists

def argparse_filestring_type(string):
    if not exists(string):
        raise argparse.ArgumentTypeError("%s is not a file." % string)
    else:
        return string

def argparse_string_type(case_converter, prefer_hyphen=False):
    if prefer_hyphen:
        return lambda string: case_converter(string).replace("_","-")
    else:
        return lambda string: case_converter(string).replace("-","_")

def split_options(opts):
    options = {}
    if opts is None:
        return options
    for o in opts:
        r = o.split('=', 1)
        options[r[0]] = None if len(r) == 1 else r[1]
    return options



# Copyright 2015 0xc0170
# Copyright (c) 2021 Chris Reed
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import logging

from ..tools_supported import ToolsSupported
from ..generate import Generator
from ..settings import ProjectSettings
from . import argparse_filestring_type, argparse_string_type, split_options

help = 'Build a project'

def run(args):
    # Export if we know how, otherwise return
    combined_projects = args.projects + args.project or ['']
    kwargs = split_options(args.options)
    generator = Generator(args.file)
    any_build_failed = False
    any_export_failed = False
    for project_name in combined_projects:
        for project in generator.generate(project_name):
            project.debug_dump = args.dump_debug_yaml
            clean_failed = False
            if args.clean and project.clean(args.tool) == -1:
                clean_failed = True # So we don't attempt to generate or build this project.
                any_build_failed = True
            if not clean_failed:
                if project.generate(args.tool, args.copy) == -1:
                    any_export_failed = True
                if project.build(args.tool, jobs=args.jobs, **kwargs) == -1:
                    any_build_failed = True
            if args.stop_on_failure and (any_build_failed or any_export_failed):
                break

    if any_build_failed or any_export_failed:
        return -1
    else:
        return 0

def setup(subparser):
    subparser.add_argument('-v', dest='verbosity', action='count', default=0,
                        help='Increase the verbosity of the output (repeat for more verbose output)')
    subparser.add_argument('-q', dest='quietness', action='count', default=0,
                        help='Decrease the verbosity of the output (repeat for less verbose output)')
    subparser.add_argument(
        "--dump-debug-yaml", action="store_true",
                        help="Write processed yaml to a <project>-dump.yaml file for debugging.")
    subparser.add_argument(
        "-f", "--file", help="YAML projects file", default='projects.yaml',
        type=argparse_filestring_type)
    subparser.add_argument(
        "-p", "--project", dest="projects", action='append', default=[], help="Name of the project to build")
    subparser.add_argument(
        "-t", "--tool", help="Build a project files for provided tool",
        type=argparse_string_type(str.lower, False), choices=list(ToolsSupported.TOOLS_DICT.keys()) + list(ToolsSupported.TOOLS_ALIAS.keys()))
    subparser.add_argument(
        "-c", "--copy", action="store_true", help="Copy all files to the exported directory")
    subparser.add_argument(
        "-k", "--clean", action="store_true", help="Clean project before building")
    subparser.add_argument(
        "-o", "--options", action="append", help="Toolchain options")
    subparser.add_argument(
        "-x", "--stop-on-failure", action="store_true", help="Stop on first failure")
    subparser.add_argument(
        "-j", "--jobs", action="store", type=int, default=1,
        help="Number of concurrent build jobs (not supported by all tools)")
    subparser.add_argument("project", nargs='*',
                        help="Specify projects to be generated and built")



#!/usr/bin/env python
# Copyright 2015 ARM Limited
#
# Licensed under the Apache License, Version 2.0
# See LICENSE file for details.

import copy

from .tool import Tool, Builder, Exporter

class GDB(Tool, Exporter, Builder):
    def __init__(self, workspace, env_settings):
        self.workspace = workspace
        self.env_settings = env_settings

    @staticmethod
    def get_toolnames():
        return ['gdb']

    @staticmethod
    def get_toolchain():
        return None

    def export_project(self):
        # for native debugging, no command files are necessary
        return None, []

    @staticmethod
    def is_supported_by_default(target):
        # does not require additional information
        return True


class ARMNoneEABIGDB(GDB):

    generated_project = {
        'path': '',
        'files': {
            'startupfile': '',
        }
    }

    def __init__(self, workspace, env_settings):
        super(ARMNoneEABIGDB, self).__init__(workspace, env_settings)

    @staticmethod
    def get_toolnames():
        return ['gdb']

    @staticmethod
    def get_toolchain():
        return None

    def _generate_file(self, port):
        generated_projects = copy.deepcopy(self.generated_project)
        expanded_dic = self.workspace.copy()

        expanded_dic['gdb_server_port'] = port

        project_path, startupfile = self.gen_file_jinja(
            'gdb.tmpl', expanded_dic, '%s.gdbstartup' % expanded_dic['name'], expanded_dic['output_dir']['path'])
        generated_projects['path'] = project_path
        generated_projects['files']['startupfile'] = startupfile
        return generated_projects

    def export_project(self):
        return self._generate_file(3333)

    def get_generated_project_files(self):
        return {'path': self.workspace['path'], 'files': [self.workspace['files']['startupfile']]}

    @staticmethod
    def is_supported_by_default(target):
        # does not require additional information
        return True

class JLinkGDB(ARMNoneEABIGDB):

    def export_project(self):
        return self._generate_file(2331)




# Copyright 2015 0xc0170
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import copy
import logging
import subprocess

from os.path import join, normpath, dirname, exists
from .tool import Tool, Exporter
from .cmake import CMake
from .gccarm import MakefileGccArm
from ..util import SOURCE_KEYS

logger = logging.getLogger('progen.tools.cmake_armclang')

class CMakeArmClang(CMake):
    def __init__(self, workspace, env_settings):
        super(CMakeArmClang, self).__init__(workspace, env_settings)
        self.logging = logging
        self.workspace['preprocess_linker_file'] = True

    @staticmethod
    def get_toolnames():
        return ['cmake_armclang']

    @staticmethod
    def get_toolchain():
        return 'armclang'

    def get_template(self):
        return 'cmakelist_armclang.tmpl'

    def get_workspace_template(self):
        return 'cmakelist_armclang_workspace.tmpl'
